#include "validation.h"

#include <random>

using namespace std;

// All interaction should go via this class to make sure we don't do invalid moves.
// It checks for:
// - printed positions are valid and still in the game
// - the team output is correctly formatted and valid
// - each value is assigned to exactly two positions (by the team)
// - return WA when the game is less than 2n-1 turns
// - stop playing after 2n-1 moves, and implicitly return AC
class Player {
  public:
	OutputValidator& v;
	const int n;

	// The current turn. 0-based.
	int turn = 0;
	// For each of 2*n positions, the value that was returned. -1 is unknown, 0..n-1 otherwise.
	vector<int> values;
	// For each of the n card values, the positions it was seen.
	vector<vector<int>> value_pos;
	// For each of the n card values, whether it's still in the game.
	vector<bool> done;

	// The number of values that's done.
	int num_done = 0;
	// The number of known/unknown (but not done) positions.
	int num_known = 0;
	int num_unknown;

	// Permutation in which the cards are turned.
	vector<int> permutation;

	Player(OutputValidator& v, int n)
	    : v(v), n(n), values(2 * n, -1), value_pos(n), done(n, false), num_unknown(2 * n),
	      permutation(2 * n) {
		for(int i = 0; i < 2 * n; ++i) { permutation[i] = i + 1; }
	}

	// The play should always reach 2n-1 turns, at which point either AC or WA
	// is given and exit() is called, skipping this destructor.
	// NOTE: This is not called in program exit from AC() and WA() functions on the output validator
	// class. However, this requires that Player is *not used as a static variable*!
	~Player() {
		cerr << "Player exited before the game was complete and before 2n-1 turns!" << endl;
		assert(false);
	}

	// True when we queried the first card in a turn, but not yet the second card.
	bool first_move_done = false;
	// If first_move_done is true, the position and value of the first move.
	int last_move = -1;
	int last_read = -1;

	// This is set to true as soon as we detect there is a winning strategy to the end of the game.
	bool forcing_win = false;

	// Given position 0<=p<2n, return a card value 0<=x<n.
	// Makes sure that the submission is consistent and that the position and card value are in the
	// right range.
	int make_move(int p) {
		assert(turn < 2 * n - 1);
		if(first_move_done) { assert(p != last_move); }

		if(not(0 <= p and p < 2 * n)) {
			cerr << "Position " << p << " is out of range!\n";
			assert(0 <= p and p < 2 * n);
		}
		if(values[p] >= 0) {
			// Make sure we do not query positions that are already done.
			if(done[values[p]]) {
				cerr << "Position " << p << " is done already!\n";
				assert(not done[values[p]]);
			}
		}
		std::cout << permutation[p] << std::endl;

		// Read the returned value at position p.
		const int x = v.read_integer("x", 1, n) - 1;
		v.newline();

		if(values[p] >= 0) {
			v.check(x == values[p], "Found inconsistent cards ", values[p] + 1, " and ", x + 1,
			        " for position ", p, "!");
		} else {
			--num_unknown;
			++num_known;
			// cerr << "READ POSITION: " << x + 1 << endl;
			value_pos[x].push_back(p);
			v.check(value_pos[x].size() <= 2, "Saw value ", x + 1, " at more than 2 positions!");
			values[p] = x;
		}

		if(first_move_done) {
			++turn;
			if(last_read == x) {
				assert(not done[x]);
				done[x] = true;
				++num_done;
				num_known -= 2;
				// cerr << "NUM DONE = " << num_done << endl;
				if(num_done == n) {
					assert(turn <= 2 * n - 1);
					v.check(turn == 2 * n - 1, "Game finished in ", turn,
					        " turns, but should have lasted at least ", 2 * n - 1, "!");
				}
			}
			last_move = -1;
			last_read = -1;
		} else {
			last_move = p;
			last_read = x;
		}

		first_move_done = !first_move_done;

		if(turn == 2 * n - 1) { v.eof_and_AC(); }

		// If we are not forcing a win yet, can we do so?
		if(not forcing_win) {
			bool can_complete_move = first_move_done and value_pos[x].size() == 2;
			try_win(can_complete_move);
		}

		return x;
	}

	// Given two 0-based positions, returns the two 0-based card values.
	pair<int, int> make_turn(int p1, int p2) {
		assert(!first_move_done);
		int x = make_move(p1);
		int y = make_move(p2);

		return {x, y};
	}

	// Check if we can force a win/WA (<2n-1 turns) at this point, and play the
	// game to the end if this is possible.
	// can_complete_move is true if only the first move of the turn is done,
	// and we can match the card that has just been turned over.
	void try_win(bool can_complete_move) {

		// First check whether we can force a win at all.
		{
			assert(num_known + num_unknown + 2 * num_done == 2 * n);
			int known = num_known, unknown = num_unknown, done = num_done;
			int turn2 = turn;
			// Deal with the 2nd move in the current turn if needed.
			if(first_move_done) {
				++turn2;
				if(can_complete_move) {
					known -= 2;
					++done;
				} else {
					++known;
					--unknown;
				}
			}
			int turns_needed;
			if(unknown == 0)
				turns_needed = known / 2;
			else if(unknown % 2 == 1)
				turns_needed = unknown / 2 + (known + unknown) / 2;
			else
				turns_needed = unknown / 2 - 1 + (known + unknown) / 2;
			// cerr << "current turn " << turn << " (" << first_move_done << " / " <<
			// can_complete_move
			//<< ")" << endl;
			// cerr << "turns_needed " << turns_needed << endl;
			// cerr << "turn2   " << turn2 << endl;
			// cerr << "known   " << known << endl;
			// cerr << "unknown " << unknown << endl;
			if(turn2 + turns_needed >= 2 * n - 1) return;
			// cerr << "VALIDATOR CAN WIN!\n";
		}

		// We can win in < 2n-1 turns, so play to completion.
		forcing_win = true;

		// List of all positions with an unknown value.
		vector<int> unknown_pos;
		for(int i = 0; i < 2 * n; ++i)
			if(values[i] == -1) {
				unknown_pos.push_back(i);
				// cerr << "unknown pos " << i + 1 << endl;
			}

		// cerr << "unknown " << num_unknown << "  size " << unknown_pos.size() << endl;

		// Handle the first halfway done turn.
		if(first_move_done) {
			if(can_complete_move) {
				assert(last_read != -1);
				// cerr << "Complete move\n";
				if(last_move == value_pos[last_read][0]) {
					make_move(value_pos[last_read][1]);
				} else {
					make_move(value_pos[last_read][0]);
				}
			} else {
				assert(not unknown_pos.empty());
				make_move(unknown_pos.back());
				unknown_pos.pop_back();
			}
		}
		// cerr << "unknown " << num_unknown << "  size " << unknown_pos.size() << endl;

		// Now, we only have full moves left.
		// We first turn over unknown cards until at most 2 unknown cards remain.
		// With at most 2 unknown cards, we can turn over an unknown card and immediately find its
		// pair. When there are no more unknown cards we pair up all known cards.

		assert(unknown_pos.size() == num_unknown);
		while(num_unknown > 2) {
			make_turn(unknown_pos[num_unknown - 2], unknown_pos.back());
			unknown_pos.pop_back();
			unknown_pos.pop_back();
		}

		while(num_unknown > 0) {
			// Query (2nd to) last one.
			int x = make_move(unknown_pos.back());
			unknown_pos.pop_back();
			if(value_pos[x].size() == 1) {
				// Match to last unknown pos.
				make_move(unknown_pos[0]);
			} else {
				// Match to already known pos.
				make_move(value_pos[x][0]);
			}
		}

		assert(num_unknown == 0);
		for(int i = 0; i < n; ++i)
			if(not done[i]) make_turn(value_pos[i][0], value_pos[i][1]);
		// By this point, we should have won and WA'd already.
		assert(false);
	}
};

// A simple player that always finishes in exactly 2n turns.
// - Query card 1 to 2n
// - Print all pairs.
void simple_player(Player& p, int n) {
	for(int i = 0; i < n; ++i) { p.make_turn(2 * i, 2 * i + 1); }
	for(int x = 0; x < n; ++x) {
		if(p.done[x]) continue;
		p.make_turn(p.value_pos[x][0], p.value_pos[x][1]);
	}
}

// A player that always uses exactly 2*n-1 turns.
// - Query card 1 to 2n-2
// - Query card 2n-1
// - Infer card 2n
// - Match card 2n-1
// - Print all pairs.
void ok_player(Player& p, int n) {
	for(int i = 0; i < 2 * n - 2; ++i) p.make_move(i);
	int x = p.make_move(2 * n - 2);
	if(p.value_pos[x].size() == 1)
		p.make_move(2 * n);
	else
		p.make_move(p.value_pos[x][0]);

	// match position 2n
	if(p.values[2 * n - 1] == -1) {
		int value = p.make_move(2 * n - 1);
		p.make_move(p.value_pos[value][0]);
	}

	// match all others
	for(int i = 0; i < n; ++i) {
		if(not p.done[i]) p.make_turn(p.value_pos[i][0], p.value_pos[i][1]);
	}
}

// A player that finishes in at most 2n-1 turns, and potentially less turns.
// - Query card 1 to 2n-2
// - Make matches as soon as possible
// - Query card 2n-1
// - Match card 2n-1
// - Print all pairs.
void optimal_player(Player& p, int n) {
	for(int i = 0; i < 2 * n - 2;) {
		int x = p.make_move(i++);
		if(p.value_pos[x].size() == 2)
			p.make_move(p.value_pos[x][0]);
		else {
			p.make_move(i++);
		}
	}
	int x = p.make_move(2 * n - 2);
	if(p.value_pos[x].size() == 2) {
		p.make_move(p.value_pos[x][0]);
	} else {
		p.make_move(2 * n - 1);
	}
	for(int x = 0; x < n; ++x) {
		if(p.done[x]) continue;
		assert(!p.value_pos[x].empty());
		if(p.value_pos[x].size() == 1) {
			p.make_turn(p.value_pos[x][0], 2 * n - 1);
		} else {
			p.make_turn(p.value_pos[x][0], p.value_pos[x][1]);
		}
	}
}

// Asks for random positions, making sure to not ask positions that are done already.
void random_player(Player& p, int n) {
	for(int i = 0; i < 2 * n - 1; ++i) {
		// Find two distinct positions for integers that aren't done yet.
		int a, b;
		do { a = rand() % (2 * n); } while(p.values[a] >= 0 and p.done[p.values[a]]);
		do { b = rand() % (2 * n); } while(b == a or (p.values[b] >= 0 and p.done[p.values[b]]));
		p.make_turn(a, b);
	}
}

// Always asks the same two positions, in random order.
void fixed_player(Player& p, int n) {
	int a = rand() % (2 * n);
	int b = rand() % (2 * n - 1);
	if(b >= a) ++b;
	while(true) {
		if(rand() % 2) swap(a, b);
		p.make_turn(a, b);
	}
}

int main(int argc, char** argv) {
	std::ifstream in(argv[1]);
	OutputValidator v(argc, argv);

	int n;
	in >> n;

	string mode;
	in >> mode;

	string perm_mode;
	in >> perm_mode;

	int seed;
	in >> seed;
	srand(seed);

	cout << n << std::endl;

	Player p(v, n);

	if(perm_mode == "decreasing")
		reverse(p.permutation.begin(), p.permutation.end());
	else if(perm_mode == "random")
		shuffle(p.permutation.begin(), p.permutation.end(), std::mt19937(rand()));
	else
		assert(perm_mode == "increasing");

	if(mode == "simple")
		simple_player(p, n);
	else if(mode == "ok")
		ok_player(p, n);
	else if(mode == "optimal")
		optimal_player(p, n);
	else if(mode == "random")
		random_player(p, n);
	else if(mode == "fixed")
		fixed_player(p, n);
	else
		assert(false);
}
